﻿//////////////////////////////////////////////////////////////////////////////
//
//  Flow chart control source  ( for PCD46x1 )
//
//  NIPPON PULSE MOTOR Co., Ltd.
//
//////////////////////////////////////////////////////////////////////////////

using FTD2XX_NET;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace MP_Builder_Sample
{
    public class FlowProcess
    {
        ////////////////////////////////////////////////////////////////////
        //Common processing

        ////////////////////////////////////////////////////////////////////
        [DllImport("winmm.dll", EntryPoint = "timeGetTime")]
        public static extern uint GetTime();

        ////////////////////////////////////////////////////////////////////
        //Work variables during acceleration/deceleration
        public bool ZeroFlag = false;  // "true" if the addition or subtraction result is zero
        public bool OverFlag = false;  // "true" when the addition/subtraction result is overflow (underflow)
        public byte[] REG = new byte[256];  // Register definition (initialization omitted)

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// register addition
        /// <summary>
        public void REG_Addition(ref byte Target, byte wb1)
        {
          int wr1 = (int)Target + (int)wb1;
          ZeroFlag = false;
          OverFlag = false;
          //overflow
          if(wr1 > 0xFF)
          {
            OverFlag = true;
          }
          //Addition result is zero
          if(wr1 == 0)
          {
            ZeroFlag = true;
          }
          Target = (byte)(wr1 & 0xFF);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// register subtraction
        /// <summary>
        public void REG_Subtraction(ref byte Target, byte wb1)
        {
          int wr1 = (int)Target - (int)wb1;
          ZeroFlag = false;
          OverFlag = false;
          //underflow
          if(wr1 < 0)
          {
            OverFlag = true;
          }
          //Subtract result is zero
          if(wr1 == 0)
          {
            ZeroFlag = true;
          }
          Target = (byte)(wr1 & 0xFF);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// WAIT
        /// <summary>
        public void Wait(Form1 fm1, uint WaitTime)
        {
            ulong StartTick = 0;
            uint NowTick = 0;
            ulong ElapsedTick = 0;

            //WAIT processing is performed after the operation of all axes is completed.
            fm1.hFlow.WaitMotionEnd(fm1, 0, fm1.hDataSmp.UseAxis);
            //
            //Start WAIT processing
            if (WaitTime == 0)
            {
                return;
            }
            //
            StartTick = (ulong)GetTime();
            ulong PreTick = StartTick;
            //
            do
            {
                NowTick = GetTime();
                if(PreTick != NowTick)
                {
                    PreTick = NowTick;
                    //
                    if (NowTick > StartTick)
                    {
                        ElapsedTick = (ulong)(NowTick - StartTick);
                    }
                    else
                    {
                        ElapsedTick = (ulong)((0x100000000 -  StartTick) + (ulong)NowTick);
                    }
                }
            } while (ElapsedTick < (ulong)WaitTime) ;
        }

        //
        //PCD46x1 unique processing ================================================================
        //

            ////////////////////////////////////////////////////////////////////
        //
        //Number of types of registers required for operation + 1 (start command)
        private const int MaxReg = 9;
        private const int ComNo = 8; // the position of the start command in the array
        //commands
        public byte[] ComVal = new byte[]  { 0x80, //RMV
                                             0x81, //RFL
                                             0x82, //RFH
                                             0x83, //RUD
                                             0x83, //dumy !!
                                             0x84, //RMG
                                             0x85, //RDP
                                             0x86, //RIDL
                                             0x87, //RENV
                                             0x88, //RCUN
                                             0x89, //RSTS (Read only)
                                             0x8A, //RIOP
                                             0x8B, //RSPO
                                             0x8C, //RSPM
                                             0x87, //RIDC
                                             0x86};//RSPD
        private const int RMV = 0;
        private const int RFL = 1;
        private const int RFH = 2;
        private const int RUD = 3;
        private const int RMG = 4;
        private const int dumy = 5;
        private const int RDP = 6;
        private const int RIDL = 7;
        private const int RENV = 8;
        private const int RCUN = 9;
        private const int RSTS = 10;
        public const int RIOP = 11;
        private const int RSPO = 12;
        private const int RSPM = 13;
        private const int RIDC = 14;
        private const int RSPD = 15;
        public int RegByte = 3;
        public int[] RegMask = new int[]  { 0x00FFFFFF, // RMV
                                            0x0000FFFF, // RFL
                                            0x0000FFFF, // RFH
                                            0x0000FFFF, // RUD
                                            0x0000FFFF, // RMG
                                            0x00FFFFFF, // RDP
                                            0x000000FF, // RIDL
                                            0x0000FFFF, // RENV
                                            0x00FFFFFF, // RCUN
                                            0x0000FFFF, // RSTS
                                            0x000000FF, // RIOP
                                            0x000000FF, // RSPO
                                            0x000000FF, // RSPM
                                            0x00FF0000, // RIDC
                                            0x00FFFF00}; // RSPD
        //Hold register value
        public uint[,] REGD = new uint[4, 14];

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// MSTS, SSTS, REGD Clear each array
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public void InitDatas(Form1 fm1)
        {
            for (int loopA = 0; loopA < 4; loopA++)
            {
                //
                fm1.hFTDI.MSTS[loopA] = 0;
                fm1.hFTDI.SSTS[loopA] = 0;
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// LSI initial settings
        /// type = 0 : initialization before operation (including excitation on by port control)
        ///        1 : Initialization before operation (including excitation off by port control)
        ///        2 : Excitation off only after operation complete
        /// </summary>
        public FTDI.FT_STATUS InitPulseGenelator(Form1 fm1, int type)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            //
            fm1.hFTDI.Read_Port(fm1);
            //
            fm1.hFTDI.index = 0;
            //
            if ((type == 0) ||
                (type == 1))
            {
                //
                for (int loops = 0, ax = 1; loops < fm1.hDataSmp.AxisConts; loops++)
                {
                    //RENV
                    fm1.hFTDI.Write_REG((byte)ax, ComVal[RENV], (uint)fm1.hDataSmp.Renv_ini[loops], RegMask[RENV]);
                    //
                    //PORT
                    PortControl(fm1, type, loops);
                    //
                    ax = ax << 1;
                }
            }
            else
            {
                for (int loops = 0, ax = 1; loops < fm1.hDataSmp.AxisConts; loops++)
                {
                    PortControl(fm1, 0, loops);
                }
            }
            //
            InitDatas(fm1);

            ftStatus = fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
            return (ftStatus);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Generic port control
        /// </summary>
        /// <param name="CheckAxis"></param>
        public void PortControl(Form1 fm1, int type, int axnum)
        {
            bool portEn = false;
            uint RPIO_D = REGD[axnum, RIOP] & 0x0000003F;
            uint num = 1;
            byte ax = (byte)(0x01 << axnum);
            for (int loopA = 0; loopA < 4; loopA++) //4 ports per axis
            {
                uint wu1 = fm1.hDataSmp.Port_set[axnum, loopA];
                if (type == 0)
                {
                    wu1 = fm1.hDataSmp.Port_reset[axnum, loopA];
                }
                //Port control commands should be the same as PSPC4.
                //Reset system is 10h、11h、12h、13h
                //Set system is 18h、19h、1Ah、1Bh
                if (wu1 != 0)
                {
                    portEn = true;
                    //
                    if ((wu1 & 0x08) == 0x08)
                    {
                        //PORT is Set
                        RPIO_D = RPIO_D | num;
                    }
                    else
                    {
                        //PORT is Reset
                        RPIO_D = RPIO_D & ~num;
                    }
                }
                //
                num = num << 1;
            }
            //
            if (portEn == true)
            {
                //RIOP
                fm1.hFTDI.Write_REG((byte)ax, ComVal[RIOP], RPIO_D, RegMask[RIOP]);
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Check the operation status of the specified axis
        ///   CheckSts：Comparison value with Status
        ///             Wait until Status is less than or equal to this value
        ///             
        ///             Return value "Status" of "hFTDI.WaitMsts" is as follows
        ///                 Status＝0：Stopped (the next operation can be written)
        ///                 Status＝1：In operation, pre-registration is free (the next operation can be written)
        ///                 Status＝2：undefined
        ///                 Status＝3：pre-register full (cannot write next operation)
        /// </summary>
        /// <param name="CheckAxis"></param>
        public void WaitMotionEnd(Form1 fm1, int CheckSts, int CheckAxis)
        {
            int EvCnt = 0;
            int Status = 3;
            while (Status > CheckSts)
            {
                EvCnt++;
                if (EvCnt > 100)
                {
                    EvCnt = 0;
                    System.Windows.Forms.Application.DoEvents();
                }
                //
                Status = fm1.hFTDI.WaitMsts(fm1, fm1.hDataSmp.AxisConts, CheckAxis);
                //
                if (fm1.hFTDI.FlowData_stop == true)//Stop processing was performed on the Form1 side
                {
                    //Write the STOP command to all axes when interrupting
                    fm1.hFTDI.index = 0;
                    fm1.hFTDI.Write_COM(0x0F, (byte)0x08);
                    fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
                    //
                    fm1.hFTDI.FlowData_on = true;
                }
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        ///  Execute positioning movement
        ///  (LSI access processing)
        /// <summary>
        public void MotioneProcess(Form1 fm1, int DataNum)
        {
            //For grouping the same start command
            byte[] StCom = new byte[] {0x10, 0x11, 0x15, 0x14};
            byte[] StAx = new byte[] {0, 0, 0, 0, 0};
            int UseAxis = 0;//Information about the axis to be executed
            //
            fm1.hFTDI.index = 0;
            //
            for (int loopA = 0, ax = 1; loopA < 4; loopA++)//4-axis
            {
                if(fm1.hDataSmp.RegDat[DataNum, loopA, ComNo] != 0)
                {
                    //Axes with no start command specified are not subject to execution
                    //
                    //Create register write data (RMV, RFL, RFH, RUR, RMG, RDP)
                    //Note that PCD46x1 does not have RDR (deceleration rate) register
                    for (int loopB = 0; loopB < (ComNo - 1); loopB++)
                    {
                        if (loopB != 4)
                        {
                            //Since the PCD46x1 does not have a deceleration register, the value corresponding to the deceleration register is basically ignored.
                            if ((loopB == 3) &&
                                (fm1.hDataSmp.RegDat[DataNum, loopA, ComNo] == 0x14))
                            {
                                //In MP_Builder, if the start command is 0x14
                                //Expect deceleration operation from FH speed to FL speed.
                                //At this time, the acceleration rate register should be zero, so substitute the value of the deceleration rate.
                                fm1.hFTDI.Write_REG((byte)ax, ComVal[loopB], (uint)fm1.hDataSmp.RegDat[DataNum, loopA, (loopB + 1)], RegMask[loopB]);
                            }
                            else
                            {
                                fm1.hFTDI.Write_REG((byte)ax, ComVal[loopB], (uint)fm1.hDataSmp.RegDat[DataNum, loopA, loopB], RegMask[loopB]);
                            }
                        }
                    }
                    //Write corresponding to RMD register
                    //PCD46x1 series contains mode commands etc. in RegDat[7] (instead of PRMD)
                    //               ↑ (1) Output mode, (2) Register select, (3) Control mode, (4) Start mode
                    //  Must write (1)(3)
                    //(3)
                    uint bcom = ((uint)fm1.hDataSmp.RegDat[DataNum, loopA, (ComNo - 1)] >> 8) & 0x0000003F | 0x00000040;
                    fm1.hFTDI.Write_COM((byte)ax, (byte)(bcom));
                    //(1)
                    bcom = ((uint)fm1.hDataSmp.RegDat[DataNum, loopA, (ComNo - 1)] >> 24) & 0x0000003F | 0x000000C0;
                    fm1.hFTDI.Write_COM((byte)ax, (byte)(bcom));

                    //
                    //group the same start commands together
                    //StAx[] holds axis information
                    switch (fm1.hDataSmp.RegDat[DataNum, loopA, ComNo])
                    {
                        case 0x10://STAFL
                            StAx[0] = (byte)(StAx[0] | ax);
                            break;
                        case 0x11://STAFH
                            StAx[1] = (byte)(StAx[1] | ax);
                            break;
                        case 0x15://STAUD
                            StAx[2] = (byte)(StAx[3] | ax);
                            break;
                        case 0x14://Instead of STAD. Implement the non-recommended FL constant speed.
                            StAx[3] = (byte)(StAx[2] | ax);
                            break;
                        default:
                            break;
                    }
                    //
                    UseAxis = UseAxis | ax;
                }
                //
                ax = ax << 1;
            }
            //
            //Create write data for start command
            for(int loopB = 0; loopB < 4; loopB++)
            {
                if(StAx[loopB] != 0)
                {
                    if (loopB < 3)
                    {
                        fm1.hFTDI.Write_COM(StAx[loopB], StCom[loopB]);
                    }
                    else
                    {
                        //Switch to "Decelerate to FL at FH start" operation
                        fm1.hFTDI.Write_COM(StAx[loopB], 0x11);//STAFH
                        fm1.hFTDI.Write_COM(StAx[loopB], 0x14);//Decelerate to FL
                    }
                }
            }
            //Wait until the LSI is ready to write the next command
            WaitMotionEnd(fm1, 0, UseAxis);
            //
            //Send command to USB
            if(fm1.hFTDI.FlowData_stop == false)
            {
                fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
            }
        }
    }
}
