﻿//////////////////////////////////////////////////////////////////////////////
//
//  Flow chart control source  ( for PCL6045 )
//
//  NIPPON PULSE MOTOR Co., Ltd.
//
//////////////////////////////////////////////////////////////////////////////

using FTD2XX_NET;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace MP_Builder_Sample
{
    public class FlowProcess
    {
        ////////////////////////////////////////////////////////////////////
        //Common processing

        ////////////////////////////////////////////////////////////////////
        [DllImport("winmm.dll", EntryPoint = "timeGetTime")]//高精度な時間待ちで使用
        public static extern uint GetTime();

        ////////////////////////////////////////////////////////////////////
        //Work variables during acceleration/deceleration
        public bool ZeroFlag = false;  // "true" if the addition or subtraction result is zero
        public bool OverFlag = false;  // "true" when the addition/subtraction result is overflow (underflow)
        public byte[] REG = new byte[256];  // Register definition (initialization omitted)

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// register addition
        /// <summary>
        public void REG_Addition(ref byte Target, byte wb1)
        {
          int wr1 = (int)Target + (int)wb1;
          ZeroFlag = false;
          OverFlag = false;
          //overflow
          if(wr1 > 0xFF)
          {
            OverFlag = true;
          }
          //Addition result is zero
          if(wr1 == 0)
          {
            ZeroFlag = true;
          }
          Target = (byte)(wr1 & 0xFF);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// register subtraction
        /// <summary>
        public void REG_Subtraction(ref byte Target, byte wb1)
        {
          int wr1 = (int)Target - (int)wb1;
          ZeroFlag = false;
          OverFlag = false;
          //underflow
          if(wr1 < 0)
          {
            OverFlag = true;
          }
          //Subtract result is zero
          if(wr1 == 0)
          {
            ZeroFlag = true;
          }
          Target = (byte)(wr1 & 0xFF);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// WAIT
        /// <summary>
        public void Wait(Form1 fm1, uint WaitTime)
        {
            ulong StartTick = 0;
            uint NowTick = 0;
            ulong ElapsedTick = 0;

            //WAIT processing is performed after the operation of all axes is completed.
            fm1.hFlow.WaitMotionEnd(fm1, 0, fm1.hDataSmp.UseAxis);
            //
            //Start WAIT processing
            if (WaitTime == 0)
            {
                return;
            }
            //
            StartTick = (ulong)GetTime();
            ulong PreTick = StartTick;
            //
            do
            {
                NowTick = GetTime();
                if(PreTick != NowTick)
                {
                    PreTick = NowTick;
                    //
                    if (NowTick > StartTick)
                    {
                        ElapsedTick = (ulong)(NowTick - StartTick);
                    }
                    else
                    {
                        ElapsedTick = (ulong)((0x100000000 -  StartTick) + (ulong)NowTick);
                    }
                }
            } while (ElapsedTick < (ulong)WaitTime) ;
        }

        //
        //PCL6045 unique processing ================================================================
        //
        ////////////////////////////////////////////////////////////////////
        //
        //Number of types of registers required for operation + 1 (start command)
        private const int MaxReg = 9;
        private const int ComNo = 8; // the position of the start command in the array
        //commands
        public byte[] ComVal = new byte[]  { 0x80, //PRMV
                                             0x81, //PRFL
                                             0x82, //PRFH
                                             0x83, //PRUR
                                             0x84, //PRDR
                                             0x85, //PRMG
                                             0x86, //PRDP
                                             0x87, //PRMD
                                             0x88, //RRIP
                                             0x8A, //RRUS
                                             0x8B, //RRDS
                                             0x9C, //RENV1
                                             0x9D, //RENV2
                                             0x9E, //RENV3
                                             0x9F, //RENV4
                                             0xA0, //RENV5
                                             0xA1, //RENV6
                                             0xA2, //RENV7
                                             0xFF};//COM
        private const int PRMV = 0;
        private const int PRFL = 1;
        private const int PRFH = 2;
        private const int PRUR = 3;
        private const int PRDR = 4;
        private const int PRMG = 5;
        private const int PRDP = 6;
        private const int PRMD = 7;
        private const int RRIP = 9;
        private const int RRUS = 9;
        private const int RRDS = 10;
        private const int RENV1 = 11;
        private const int RENV2 = 12;
        private const int RENV3 = 13;
        private const int RENV4 = 14;
        private const int RENV5 = 15;
        private const int RENV6 = 16;
        private const int RENV7 = 17;
        public int RegByte = 4;
        public int[] RegTble = new int[]  {  0,  // RMV
                                             1,  // RFL
                                             2,  // RFH
                                             3,  // RUR
                                             4,  // RDR
                                             5,  // RMG
                                             6,  // RDP
                                             7,  // RMD
                                            19}; // RCOM

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// MSTS, SSTS, REGD Clear each array
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public void InitDatas(Form1 fm1)
        {
            for (int loopA = 0; loopA < 4; loopA++)
            {
                //
                fm1.hFTDI.MSTS[loopA] = 0;
                fm1.hFTDI.SSTS[loopA] = 0;
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// LSI initial settings
        /// type = 0 : initialization before operation (including excitation on by port control)
        ///        1 : Initialization before operation (including excitation off by port control)
        ///        2 : Excitation off only after operation complete
        /// </summary>
        public FTDI.FT_STATUS InitPulseGenelator(Form1 fm1, int type)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            //
            fm1.hFTDI.index = 0;
            //
            if ((type == 0) ||
                (type == 1))
            {
                //
                for (int loops = 0, ax = 1; loops < fm1.hDataSmp.AxisConts; loops++)
                {
                    //RENV1
                    fm1.hFTDI.Write_REG((byte)ax, ComVal[RENV1], fm1.hDataSmp.Renv1_ini[loops]);
                    //RENV2
                    fm1.hFTDI.Write_REG((byte)ax, ComVal[RENV2], fm1.hDataSmp.Renv2_ini[loops]);
                    //RENV3
                    fm1.hFTDI.Write_REG((byte)ax, ComVal[RENV3], fm1.hDataSmp.Renv3_ini[loops]);
                    //RENV4
                    fm1.hFTDI.Write_REG((byte)ax, ComVal[RENV4], fm1.hDataSmp.Renv4_ini[loops]);
                    //
                    //PORT
                    if (type == 0)
                    {
                        for (int loopA = 0; loopA < 8; loopA++)
                        {
                            if (fm1.hDataSmp.Port_set[loops, loopA] != 0)
                            {
                                fm1.hFTDI.Write_COM((byte)ax, (byte)(fm1.hDataSmp.Port_set[loops, loopA]));//PORT-n
                            }
                        }
                    }
                    else
                    {
                        for (int loopA = 0; loopA < 8; loopA++)
                        {
                            if (fm1.hDataSmp.Port_reset[loops, loopA] != 0)
                            {
                                fm1.hFTDI.Write_COM((byte)ax, (byte)(fm1.hDataSmp.Port_reset[loops, loopA]));//PORT-n
                            }
                        }
                    }
                    //
                    ax = ax << 1;
                }
            }
            else
            {
                for (int loops = 0, ax = 1; loops < fm1.hDataSmp.AxisConts; loops++)
                {
                    for (int loopA = 0; loopA < 8; loopA++)
                    {
                        if (fm1.hDataSmp.Port_reset[loops, loopA] != 0)
                        {
                            fm1.hFTDI.Write_COM((byte)ax, (byte)(fm1.hDataSmp.Port_reset[loops, loopA]));//PORT-n
                        }
                    }
                }
            }
            //
            InitDatas(fm1);

            ftStatus = fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
            return (ftStatus);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// 指定した軸の動作状況を確認
        /// Check the operation status of the specified axis
        ///   CheckSts：Comparison value with Status
        ///             Wait until Status is less than or equal to this value
        ///             
        ///             Return value "Status" of "hFTDI.WaitMsts" is as follows
        ///                 Status＝0：Stopped (the next operation can be written)
        ///                 Status＝1：In operation, pre-registration is free (the next operation can be written)
        ///                 Status＝2：undefined
        ///                 Status＝3：pre-register full (cannot write next operation)
        /// </summary>
        /// <param name="CheckAxis"></param>
        public void WaitMotionEnd(Form1 fm1, int CheckSts, int CheckAxis)
        {
            int EvCnt = 0;
            int Status = 3;
            while (Status > CheckSts)
            {
                EvCnt++;
                if (EvCnt > 100)
                {
                    EvCnt = 0;
                    System.Windows.Forms.Application.DoEvents();
                }
                //
                Status = fm1.hFTDI.WaitMsts(fm1, fm1.hDataSmp.AxisConts, CheckAxis);
                //
                if (fm1.hFTDI.FlowData_stop == true)//Stop processing was performed on the Form1 side
                {
                    //Write the STOP command to all axes when interrupting
                    fm1.hFTDI.index = 0;
                    fm1.hFTDI.Write_COM(0x0F, (byte)0x49);
                    fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
                    //
                    fm1.hFTDI.FlowData_on = true;
                }
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        ///  Execute positioning movement
        ///  (LSI access processing)
        /// <summary>
        public void MotioneProcess(Form1 fm1, int DataNum)
        {
            //For grouping the same start command
            byte[] StCom = new byte[] {0x50, 0x51, 0x52, 0x53};
            byte[] StAx = new byte[] {0, 0, 0, 0};
            int UseAxis = 0;//Information about the axis to be executed
            //
            fm1.hFTDI.index = 0;
            //
            for (int loopA = 0, ax = 1; loopA < 4; loopA++)//4-axis
            {
                if(fm1.hDataSmp.RegDat[DataNum, loopA, ComNo] != 0)
                {
                    //Axes with no start command specified are not subject to execution
                    //
                    //Create register write data
                    for(int loopB = 0; loopB < ComNo; loopB++)
                    {
                        fm1.hFTDI.Write_REG((byte)ax, ComVal[loopB], fm1.hDataSmp.RegDat[DataNum, loopA, loopB]);
                    }
                    //
                    //group the same start commands together
                    //StAx[] holds axis information
                    switch(fm1.hDataSmp.RegDat[DataNum, loopA, ComNo])
                    {
                        case 0x50://STAFL
                            StAx[0] = (byte)(StAx[0] | ax);
                            break;
                        case 0x51://STAFH
                            StAx[1] = (byte)(StAx[1] | ax);
                            break;
                        case 0x52://STAD
                            StAx[2] = (byte)(StAx[2] | ax);
                            break;
                        case 0x53://STAUD
                            StAx[3] = (byte)(StAx[3] | ax);
                            break;
                    }
                    //
                    UseAxis = UseAxis | ax;
                }
                //
                ax = ax << 1;
            }
            //Create write data for start command
            for(int loopB = 0; loopB < 4; loopB++)
            {
                if(StAx[loopB] != 0)
                {
                    fm1.hFTDI.Write_COM(StAx[loopB], StCom[loopB]);
                }
            }
            //Wait until the LSI is ready to write the next command
            WaitMotionEnd(fm1, 1, UseAxis);
            //
            //Send command to USB
            if(fm1.hFTDI.FlowData_stop == false)
            {
                fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
            }
        }
    }
}
